/*
	(c) 2004-2005, Michael Lotz, mmlr@mlotz.ch
	Corrects timing issues with CPU's faster than 2.1GHz
	and corrects there displayed speed.
	
	Based on the idea of vmware_timer by François Revol.
	Includes CPU detection from Haiku/NewOS kernel.
	
	This file is licensed under the MIT.
*/

#include <Drivers.h>
#include <driver_settings.h>
#include <printf.h>

// variables
extern uint32 cv_factor;
extern uint32 info_cpu;

// kernel functions
extern get_cpu_info();

// our functions
extern uint64	rdtsc();
extern void		patch_get_cpu_info();
extern void		reexport_cv_factor();

int32 api_version = B_CUR_DRIVER_API_VERSION;

#define TIMER_CLKNUM_HZ (14318180 / 12)
#define out8(value, port) asm volatile ("outb %%al,%%dx" : : "a" (value), "d" (port))
#define in8(port) ({ uint8 _v; asm volatile ("inb %%dx,%%al" : "=a" (_v) : "d" (port)); _v; })

status_t init_hardware(void)
{
	return B_OK;
}

/* start Haiku extract */

uint32
calculate_cpu_conversion_factor()
{
	uint32 s_low, s_high;
	uint32 low, high;
	uint32 expired;
	uint32 result;
	uint64 t1, t2;
	uint64 p1, p2, p3;
	double r1, r2, r3;
	
	disable_interrupts();
	
	out8(0x34, 0x43);	/* program the timer to count down mode */
	out8(0xff, 0x40);	/* low and then high */
	out8(0xff, 0x40);

	/* quick sample */
quick_sample:
	do {
		out8(0x00, 0x43); /* latch counter value */
		s_low = in8(0x40);
		s_high = in8(0x40);
	} while(s_high != 255);
	t1 = rdtsc();
	do {
		out8(0x00, 0x43); /* latch counter value */
		low = in8(0x40);
		high = in8(0x40);
	} while (high > 224);
	t2 = rdtsc();

	p1 = t2-t1;
	r1 = (double)(p1) / (double)(((s_high << 8) | s_low) - ((high << 8) | low));

	/* not so quick sample */
not_so_quick_sample:
	do {
		out8(0x00, 0x43); /* latch counter value */
		s_low = in8(0x40);
		s_high = in8(0x40);
	} while (s_high!= 255);
	t1 = rdtsc();
	do {
		out8(0x00, 0x43); /* latch counter value */
		low = in8(0x40);
		high = in8(0x40);
	} while (high> 192);
	t2 = rdtsc();
	p2 = t2-t1;
	r2 = (double)(p2) / (double)(((s_high << 8) | s_low) - ((high << 8) | low));
	if ((r1/r2) > 1.01) {
		//dprintf("Tuning loop(1)\n");
		goto quick_sample;
	}
	if ((r1/r2) < 0.99) {
		//dprintf("Tuning loop(1)\n");
		goto quick_sample;
	}

	/* slow sample */
	do {
		out8(0x00, 0x43); /* latch counter value */
		s_low = in8(0x40);
		s_high = in8(0x40);
	} while (s_high!= 255);
	t1 = rdtsc();
	do {
		out8(0x00, 0x43); /* latch counter value */
		low = in8(0x40);
		high = in8(0x40);
	} while (high > 128);
	t2 = rdtsc();

	p3 = t2-t1;
	r3 = (double)(p3) / (double)(((s_high << 8) | s_low) - ((high << 8) | low));
	if ((r2/r3) > 1.01) {
		//dprintf("Tuning loop(2)\n");
		goto not_so_quick_sample;
	}
	if ((r2/r3) < 0.99) {
		//dprintf("Tuning loop(2)\n");
		goto not_so_quick_sample;
	}

	expired = ((s_high << 8) | s_low) - ((high << 8) | low);
	p3 *= TIMER_CLKNUM_HZ;
	{
		unsigned i;
		unsigned k;
		uint64 C;
		uint64 x0;
		uint64 x1;
		uint64 a, b, c;

		/* first calculate k*/
		k = 0;
		for (i = 12; i < 16; i++) {
			if (expired & (1<<i))
				k = i - 11;
		}

		C = 1000000ULL << 32;
		x0 = expired >> k;
		x1 = expired & ((1 << k) - 1);

		a = ((C * x0) / p3) << k;
		b = (((C * x0) % p3) << k) / p3;
		c = (C * x1) / p3;
		
		result = a + b + c;
	}

	if (p3 / expired / 1000000000LL)
		dprintf("cpu_fix: CPU at %Ld.%03Ld GHz\n", p3 / expired / 1000000000LL, ((p3 / expired) % 1000000000LL) / 1000000LL);
	else
		dprintf("cpu_fix: CPU at %Ld.%03Ld MHz\n", p3 / expired / 1000000LL, ((p3 / expired) % 1000000LL) / 1000LL);

	enable_interrupts();
	return result;
}

/* end Haiku extract */

status_t init_driver(void)
{
	void *handle;
	bool disable = false;
	bool disable_cv = false;
	bool disable_gc = false;
	
	handle = load_driver_settings("cpu_fix_settings");
	if (handle) {
		disable = get_driver_boolean_parameter(handle, "disable", false, true);
		disable_cv = get_driver_boolean_parameter(handle, "disable_cv_factor", false, true);
		disable_gc = get_driver_boolean_parameter(handle, "disable_get_cpu_info", false, true);
	}
	unload_driver_settings(handle);
	
	if (disable)
		return B_ERROR;
		
	if (!disable_gc) {
		dprintf("cpu_fix: patching get_cpu_info...\n");
		memcpy((uint8 *)get_cpu_info, (uint8 *)patch_get_cpu_info, 47);
	}
	
	if (!disable_cv) {
		double temp;
		int64 tsc;
		int32 new_cv_factor;

		dprintf("cpu_fix: fixing cv_factor...\n");

		// when changing the cv_factor we also have to change the system_time_base.
		// otherwise the calculation based on the new cv_factor will be larger/smaller
		// and we jump in time (uptime with new cv_factor would be different).
		new_cv_factor = calculate_cpu_conversion_factor();
		tsc = rdtsc();
		temp = tsc;
		temp += system_time_base;
		temp *= cv_factor;
		temp /= new_cv_factor;
		temp -= tsc;
		system_time_base = (int64)temp;
		cv_factor = new_cv_factor;

		reexport_cv_factor();
	}
	
	return B_ERROR; // we do not really support any devices
}

void uninit_driver(void)
{
}

const char **publish_devices()
{
	return NULL;
}

device_hooks *find_device(const char *name)
{
	return NULL;
}
